"use strict";

var max_M = 32;

window.onload = function init()
{
  var canvas = document.getElementById("webgl");

  var gl = WebGLUtils.setupWebGL(canvas);
  if (!gl) {
    console.log('Failed to get the rendering context for WebGL');
    return;
  }
  // We need an extension to use floating point textures
  var ext = gl.getExtension('OES_texture_float');
  if (!ext) {
    console.log('Warning: Unable to use an extension, some functionality may not work as intended');
  }
  // We need an extension to call dFdx and dFdy in the fragment shader
  var ext = gl.getExtension('OES_standard_derivatives');
  if (!ext) {
    console.log('Warning: Unable to use an extension, some functionality may not work as intended');
  }

  gl.viewport(0, 0, canvas.width, canvas.height);
  gl.clearColor(1.0, 1.0, 1.0, 1.0);
  gl.enable(gl.DEPTH_TEST);
  gl.enable(gl.CULL_FACE);

  // Load shaders and initialize attribute buffers
  gl.program = initShaders(gl, "LM-vertex-shader", "LM-fragment-shader");
  gl.useProgram(gl.program);
  initTexture(gl);
  initZeroArgLegendreTexture(gl);

  // Set the projection matrix
  gl.mvpMatrix = gl.getUniformLocation(gl.program, "mvpMatrix");
  gl.vMatrix = gl.getUniformLocation(gl.program, "vMatrix");
  var projection = perspective(45.0, canvas.width/canvas.height, 1.0e-2, 1.0e3);

  var subdivs = 7;
  var n = initVertexBuffers(gl, subdivs);

  // Setup the view by specifying eye and lookat
  var eye = vec3(0.0, 0.0, 1.5);
  var lookat = vec3(0.0, 0.0, 0.0);
  var up = vec3(0.0, 1.0, 0.0);
  //var view = lookAt(eye, lookat, up);

  // Initialize trackball
  var z_dir = vec3(eye[0] - lookat[0], eye[1] - lookat[1], eye[2] - lookat[2]);
  var eye_dist_pan = vec3(length(z_dir), -0.65, 0.0);
  var qrot = new Quaternion();
  var qinc = new Quaternion();
  qrot = qrot.make_rot_vec2vec(vec3(0.0, 0.0, 1.0), z_dir);
  var qrot_inv = new Quaternion(qrot);
  qrot_inv.invert();
  up = qrot_inv.apply(up);

  // Register the event handler
  initEventHandlers(canvas, qrot, qinc, eye_dist_pan);

  document.getElementById("configure").onclick = function () { initTexture(gl); }
  document.getElementById("IncSubdiv").onclick = function () { n = initVertexBuffers(gl, ++subdivs); };
  document.getElementById("DecSubdiv").onclick = function () { n = initVertexBuffers(gl, --subdivs); };
  document.onkeypress = function (ev) { // Key is pressed
    switch (ev.keyCode) {
      case 13: // enter
        initTexture(gl);
        document.body.focus();
        break;
    }
  };
  document.getElementById("LMselect").onclick = function () {
    var LMtheory = parseInt(document.querySelector('input[name = "LMtheory"]:checked').value);
    gl.uniform1i(gl.getUniformLocation(gl.program, "LMtheory"), LMtheory);
  }

  var tick = function () {
    qrot = qrot.multiply(qinc);
    render(gl, n, projection, eye_dist_pan, lookat, up, qrot);
    requestAnimFrame(tick);
  };
  tick();
}

function render(gl, numVertices, projection, eye_dist_pan, lookat, up, qrot)
{   
  // Calculate the model view projection matrix and pass it to u_MvpMatrix
  var u = qrot.apply(up);
  var r = qrot.apply(vec3([1, 0, 0]));
  var p = lookat;
  var e = eye_dist_pan;
  var c = vec3([p[0] - r[0]*e[1] - u[0]*e[2], p[1] - r[1]*e[1] - u[1]*e[2], p[2] - r[2]*e[1] - u[2]*e[2]]);
  var re = qrot.apply(vec3([0, 0, e[0]]));
  var view = lookAt(vec3(re[0] + c[0], re[1] + c[1], re[2] + c[2]), c, u);
  var ctm = mult(projection, view);
  gl.uniformMatrix4fv(gl.vMatrix, false, flatten(view));
  gl.uniformMatrix4fv(gl.mvpMatrix, false, flatten(ctm));

  gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
  gl.drawArrays(gl.TRIANGLES, 0, numVertices);
}

function initTexture(gl) {
  var wavelength = parseFloat(document.getElementById("wavelength").value);
  var radius = parseFloat(document.getElementById("radius").value);
  var n_p = vec2(parseFloat(document.getElementById("n_p_re").value),
                 parseFloat(document.getElementById("n_p_im").value));
  var n_med = vec2(parseFloat(document.getElementById("n_med_re").value),
                   parseFloat(document.getElementById("n_med_im").value));
  var pol = vec2(parseFloat(document.getElementById("pol_s").value),
                 parseFloat(document.getElementById("pol_p").value));
  var alpha = vec2(parseFloat(document.getElementById("alpha_re").value),
                   parseFloat(document.getElementById("alpha_im").value));
  gl.uniform1f(gl.getUniformLocation(gl.program, "wavelength"), wavelength);
  gl.uniform1f(gl.getUniformLocation(gl.program, "radius"), radius);
  gl.uniform2f(gl.getUniformLocation(gl.program, "n_p"), n_p[0], n_p[1]);
  gl.uniform2f(gl.getUniformLocation(gl.program, "n_med"), n_med[0], n_med[1]);
  gl.uniform2f(gl.getUniformLocation(gl.program, "pol"), pol[0], pol[1]);
  gl.uniform2f(gl.getUniformLocation(gl.program, "alpha"), alpha[0], alpha[1]);
  gl.uniform1f(gl.getUniformLocation(gl.program, "ab_res"), max_M);
  gl.uniform1i(gl.getUniformLocation(gl.program, "LorenzMie_ab"), 0);

  var psize = 2.0*Math.PI*radius/wavelength;
  var tts = terms_to_sum(vec2(n_med[0]*psize, n_med[1]*psize));
  gl.uniform1i(gl.getUniformLocation(gl.program, "terms_to_sum"), tts);

  var LMtheory = parseInt(document.querySelector('input[name = "LMtheory"]:checked').value);
  gl.uniform1i(gl.getUniformLocation(gl.program, "LMtheory"), LMtheory);

  gl.activeTexture(gl.TEXTURE0);
  if(!gl.texture)
    gl.texture = gl.createTexture();
  gl.bindTexture(gl.TEXTURE_2D, gl.texture);

  var ab = LorenzMie_coefs(wavelength, radius, n_p, n_med);
  var data = new Float32Array(Math.max(max_M*4, ab.length));
  data.fill(0.0);
  data.set(ab);
  gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, max_M, 1, 0, gl.RGBA, gl.FLOAT, data);
  gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
  gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
}

function initZeroArgLegendreTexture(gl)
{
  var size = (max_M + 1)*(max_M + 1);
  var texSize = size;
  var p2counter = 0;
  while(texSize)
  {
    texSize = texSize>>>1;
    ++p2counter;
  }
  texSize = 1<<p2counter;

  gl.uniform1f(gl.getUniformLocation(gl.program, "P0_res"), texSize);
  gl.uniform1i(gl.getUniformLocation(gl.program, "P0_tex"), 1);

  gl.activeTexture(gl.TEXTURE1);
  if(!gl.p0texture)
    gl.p0texture = gl.createTexture();
  gl.bindTexture(gl.TEXTURE_2D, gl.p0texture);

  var P0 = new Float32Array(texSize);
  P0.fill(0.0);
  P0[0] = 1;
  for(var n = 1; n < max_M; ++n)
  {
    P0[n*(n + 2)] = -(n + n - 1)*P0[n*n - 1];
    var idx = n*(n + 1);
    for(var m = n - 1; m > -n; m -= 2)
      P0[idx + m - 1] = P0[idx + m + 1]/((n + m)*(m - n - 1));
  }
  gl.texImage2D(gl.TEXTURE_2D, 0, gl.LUMINANCE, texSize, 1, 0, gl.LUMINANCE, gl.FLOAT, P0);
  gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
  gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
}

function initVertexBuffers(gl, subdivs)
{
  var va = vec4(0.0, 0.0, 1.0, 1.0);
  var vb = vec4(0.0, Math.sqrt(8.0)/3.0, -1.0/3.0, 1.0);
  var vc = vec4(-Math.sqrt(6.0)/3.0, -Math.sqrt(2.0)/3.0, -1.0/3.0, 1.0);
  var vd = vec4(Math.sqrt(6.0)/3.0, -Math.sqrt(2.0)/3.0, -1.0/3.0, 1.0);
  var verts = [];
  tetraSubdiv(verts, va, vb, vc, vd, subdivs);

  var vBuffer = gl.createBuffer();
  gl.bindBuffer(gl.ARRAY_BUFFER, vBuffer);
  gl.bufferData(gl.ARRAY_BUFFER, flatten(verts), gl.STATIC_DRAW);

  var vPosition = gl.getAttribLocation(gl.program, "vPosition");
  gl.vertexAttribPointer(vPosition, 4, gl.FLOAT, false, 0, 0);
  gl.enableVertexAttribArray(vPosition);

  return verts.length;
}

function triangle(verts, a, b, c)
{
  verts.push(a);
  verts.push(b);
  verts.push(c);
}

function divideTriangle(verts, a, b, c, count)
{
  if (count > 0) {
    var ab = normalize(mix(a, b, 0.5), true);
    var ac = normalize(mix(a, c, 0.5), true);
    var bc = normalize(mix(b, c, 0.5), true);

    divideTriangle(verts, a, ab, ac, count - 1);
    divideTriangle(verts, ab, b, bc, count - 1);
    divideTriangle(verts, bc, c, ac, count - 1);
    divideTriangle(verts, ab, bc, ac, count - 1);
  }
  else
    triangle(verts, a, b, c);
}

function tetraSubdiv(verts, a, b, c, d, n)
{
  divideTriangle(verts, a, b, c, n);
  divideTriangle(verts, d, c, b, n);
  divideTriangle(verts, a, d, b, n);
  divideTriangle(verts, a, c, d, n);
}

// Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet
// if we are away from the center of the sphere.
function project_to_sphere(x, y) {
  var r = 2;
  var d = Math.sqrt(x * x + y * y);
  var t = r * Math.sqrt(2);
  var z;
  if(d < r) // Inside sphere
    z = Math.sqrt(r * r - d * d);
  else if(d < t)
    z = 0;
  else       // On hyperbola
    z = t * t / d;
  return z;
}

function initEventHandlers(canvas, qrot, qinc, eye_dist_pan) {
  var dragging = false;         // Dragging or not
  var lastX = -1, lastY = -1;   // Last position of the mouse
  var current_action = 0;       // Actions: 0 - none, 1 - orbit, 2 - zoom, 3 - pan

  canvas.onmousedown = function(ev) {   // Mouse is pressed
    ev.preventDefault();
    var x = ev.clientX, y = ev.clientY;
    // Start dragging if a mouse is in <canvas>
    var rect = ev.target.getBoundingClientRect();
    if(rect.left <= x && x < rect.right && rect.top <= y && y < rect.bottom) {
      lastX = x; lastY = y;
      dragging = true;
      current_action = ev.button + 1;
    }
  };

  canvas.oncontextmenu = function(ev) { ev.preventDefault(); };

  canvas.onmouseup = function(ev) {
    var x = ev.clientX, y = ev.clientY;
    if(x === lastX && y === lastY) {
      qinc.setIdentity();
    }
    dragging = false;
    current_action = 0;
  }; // Mouse is released

  var g_last = Date.now();
  canvas.onmousemove = function(ev) { // Mouse is moved
    var x = ev.clientX, y = ev.clientY;
    if(dragging) {
      var now = Date.now();
      var elapsed = now - g_last;
      if(elapsed > 20) {
        g_last = now;
        var rect = ev.target.getBoundingClientRect();
        var s_x = ((x - rect.left)/rect.width - 0.5)*2;
        var s_y = (0.5 - (y - rect.top)/rect.height)*2;
        var s_last_x = ((lastX - rect.left)/rect.width - 0.5)*2;
        var s_last_y = (0.5 - (lastY - rect.top)/rect.height)*2;
        switch(current_action) {
          case 1: { // orbit
            var v1 = vec3(s_x, s_y, project_to_sphere(s_x, s_y));
            var v2 = vec3(s_last_x, s_last_y, project_to_sphere(s_last_x, s_last_y));
            qinc = qinc.make_rot_vec2vec(normalize(v1), normalize(v2));
          }
            break;
          case 2: { // zoom
            eye_dist_pan[0] += (s_y - s_last_y)*eye_dist_pan[0]*0.5;
            eye_dist_pan[0] = Math.max(eye_dist_pan[0], 0.01);
          }
            break;
          case 3: { // pan
            eye_dist_pan[1] += (s_x - s_last_x)*eye_dist_pan[0]*0.25;
            eye_dist_pan[2] += (s_y - s_last_y)*eye_dist_pan[0]*0.25;
          }
            break;
        }
        lastX = x, lastY = y;
      }
    }
  };
  
  canvas.addEventListener("touchstart", startZoom, false);
  function startZoom(ev) {
    ev.preventDefault();
    if (ev.targetTouches.length === 1) {
      var touch = ev.targetTouches[0];
      touch.preventDefault = function () { };
      touch.button = 0;
      canvas.onmousedown(touch);
      this.addEventListener("touchmove", roll, false);
      this.addEventListener("touchend", release, false);

      function roll(e) {
        touch = e.targetTouches[0];
        canvas.onmousemove(touch);
      }
      function release() {
        canvas.onmouseup(touch);
        this.removeEventListener("touchmove", roll);
        this.removeEventListener("touchend", release);
      }
    }
    else if (ev.targetTouches.length === 2) {
      qinc.setIdentity();
      var touch1 = ev.targetTouches[0];
      var touch2 = ev.targetTouches[1];
      touch1.preventDefault = function () { };
      touch1.button = 2;
      canvas.onmousedown(touch1);
      var dX = touch2.clientX - touch1.clientX;
      var dY = touch2.clientY - touch1.clientY;
      var startDist = Math.sqrt(dX * dX + dY * dY);
      this.addEventListener("touchmove", zoom_pan, false);
      this.addEventListener("touchend", end, false);

      function zoom_pan(e) {
        var t1 = e.targetTouches[0];
        var t2 = e.targetTouches[1];
        var v1X = touch1.clientX - t1.clientX;
        var v1Y = touch1.clientY - t1.clientY;
        var v2X = touch2.clientX - t2.clientX;
        var v2Y = touch2.clientY - t2.clientY;
        var v1_len = Math.sqrt(v1X * v1X + v1Y * v1Y);
        var v2_len = Math.sqrt(v2X * v2X + v2Y * v2Y);
        if (v1_len < 0.01 || v2_len < 0.01) return;
        var cos_v1v2 = (v1X * v2X + v1Y * v2Y) / (v1_len * v2_len);
        var dX = t2.clientX - t1.clientX;
        var dY = t2.clientY - t1.clientY;
        var newDist = Math.sqrt(dX * dX + dY * dY);
        var delta = newDist - startDist;
        startDist = newDist;
        if (cos_v1v2 > 0.8) {
          current_action = 3;
          touch1 = t1;
          touch2 = t2;
          canvas.onmousemove(touch1);
        }
        else {
          lastY -= delta;
          current_action = 2;
          canvas.onmousemove(touch1);
        }
      }
      function end() {
        canvas.onmouseup(touch1);
        this.removeEventListener("touchmove", zoom_pan);
        this.removeEventListener("touchend", end);
      }
    }
  }
}
